# Copyright (c) 2010 OpenStack Foundation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Neutron documentation build configuration file

import logging
import os
import sys

# NOTE(amotoki): In case of oslo_config.sphinxext is enabled,
# when resolving automodule neutron.tests.functional.db.test_migrations,
# sphinx accesses tests/functional/__init__.py is processed,
# eventlet.monkey_patch() is called and monkey_patch() tries to access
# pyroute2.common.__class__ attribute. It raises pyroute2 warning and
# it causes sphinx build failure due to warning-is-error = 1.
# To pass sphinx build, ignore pyroute2 warning explicitly.
logging.getLogger('pyroute2').setLevel(logging.ERROR)

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
BASE_DIR = os.path.dirname(os.path.abspath(__file__))
NEUTRON_DIR = os.path.abspath(os.path.join(BASE_DIR, "..", ".."))
sys.path.insert(0, NEUTRON_DIR)

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.coverage',
    'sphinx.ext.ifconfig',
    'sphinx.ext.graphviz',
    'sphinx.ext.todo',
    'openstackdocstheme',
    'sphinx_feature_classification.support_matrix',
    'oslo_config.sphinxext',
    'oslo_config.sphinxconfiggen',
    'oslo_policy.sphinxext',
    'oslo_policy.sphinxpolicygen',
]

# Project cross-reference roles
openstackdocs_projects = [
    'neutron',
    'nova',
]

# openstackdocstheme options
openstackdocs_repo_name = 'openstack/neutron'
openstackdocs_pdf_link = True
openstackdocs_bug_project = 'neutron'
openstackdocs_bug_tag = 'doc'
openstackdocs_auto_name = False

todo_include_todos = True

# The suffix of source filenames.
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = 'Neutron'
copyright = '2011-present, OpenStack Foundation.'

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
show_authors = True

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'native'

# A list of ignored prefixes for module index sorting.
modindex_common_prefix = ['neutron.']

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  Major themes that come with
# Sphinx are currently 'default' and 'sphinxdoc'.
# html_theme_path = ["."]
# html_theme = '_theme'
html_theme = 'openstackdocs'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']


# -- Options for LaTeX output ------------------------------------------------

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author,
# documentclass [howto/manual]).
latex_documents = [
    ('pdf-index', 'doc-neutron.tex', 'Neutron Documentation',
     'Neutron development team', 'manual'),
]

# Disable usage of xindy https://bugzilla.redhat.com/show_bug.cgi?id=1643664
latex_use_xindy = False

latex_domain_indices = False

latex_elements = {
    'makeindex': '',
    'printindex': '',
    'preamble': r'\setcounter{tocdepth}{3}',
}

# -- Options for oslo_config.sphinxconfiggen ---------------------------------

_config_generator_config_files = [
    'dhcp_agent.ini',
    'l3_agent.ini',
    'macvtap_agent.ini',
    'metadata_agent.ini',
    'metering_agent.ini',
    'ml2_conf.ini',
    'neutron.conf',
    'openvswitch_agent.ini',
    'sriov_agent.ini',
]


def _get_config_generator_config_definition(conf):
    config_file_path = '../../etc/oslo-config-generator/%s' % conf
    # oslo_config.sphinxconfiggen appends '.conf.sample' to the filename,
    # strip file extentension (.conf or .ini).
    output_file_path = '_static/config-samples/%s' % conf.rsplit('.', 1)[0]
    return (config_file_path, output_file_path)


config_generator_config_file = [
    _get_config_generator_config_definition(conf)
    for conf in _config_generator_config_files
]

# -- Options for oslo_policy.sphinxpolicygen ---------------------------------

policy_generator_config_file = '../../etc/oslo-policy-generator/policy.conf'
sample_policy_basename = '_static/neutron'

linkcheck_anchors_ignore = [
    # skip gerrit anchors
    r'\/q\/.*',
    r'q\,.*',
    r'\/c\/.*'
]



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

import abc
import copy
import hashlib
import secrets

from oslo_log import log

from neutron.common.ovn import constants as ovn_const
from neutron.common.ovn import utils
from neutron.conf.plugins.ml2.drivers.ovn import ovn_conf


LOG = log.getLogger(__name__)


class OVNGatewayScheduler(metaclass=abc.ABCMeta):

    def __init__(self):
        pass

    @abc.abstractmethod
    def select(self, nb_idl, sb_idl, gateway_name, candidates=None,
               existing_chassis=None, target_lrouter=None):
        """Schedule the gateway port of a router to an OVN chassis.

        Schedule the gateway router port only if it is not already
        scheduled.
        """

    @staticmethod
    def filter_existing_chassis(gw_chassis, physnet, chassis_physnets,
                                existing_chassis, az_hints, chassis_with_azs):
        chassis_list = copy.copy(existing_chassis)
        for chassis_name in existing_chassis or []:
            if utils.is_gateway_chassis_invalid(chassis_name, gw_chassis,
                                                physnet, chassis_physnets,
                                                az_hints, chassis_with_azs):
                LOG.debug("Chassis %(chassis)s is invalid for scheduling "
                          "router in physnet: %(physnet)s.",
                          {'chassis': chassis_name,
                           'physnet': physnet})
                chassis_list.remove(chassis_name)
        return chassis_list

    def _schedule_gateway(self, nb_idl, sb_idl, gateway_name, candidates,
                          existing_chassis, target_lrouter):
        existing_chassis = existing_chassis or []
        candidates = candidates or []
        candidates = list(set(candidates) - set(existing_chassis))
        # If no candidates, or gateway scheduled on MAX_GW_CHASSIS nodes
        # or all candidates in existing_chassis, return existing_chassis.
        # Otherwise, if more candidates present, then schedule them.
        if existing_chassis:
            if not candidates or (
                    len(existing_chassis) == ovn_const.MAX_GW_CHASSIS):
                return existing_chassis
        if not candidates:
            LOG.warning('Gateway %s was not scheduled on any chassis, no '
                        'candidates are available', gateway_name)
            return
        chassis_count = min(
            ovn_const.MAX_GW_CHASSIS - len(existing_chassis),
            len(candidates)
        )
        # The actual binding of the gateway to a chassis via the options
        # column or gateway_chassis column in the OVN_Northbound is done
        # by the caller
        chassis = self._select_gateway_chassis(
            nb_idl, sb_idl, candidates, 1, chassis_count, target_lrouter,
            gateway_name=gateway_name)[:chassis_count]
        # priority of existing chassis is higher than candidates
        chassis = existing_chassis + chassis

        LOG.debug("Gateway %s scheduled on chassis %s",
                  gateway_name, chassis)
        return chassis

    def _reorder_by_az(self, nb_idl, sb_idl, candidates):
        chassis_selected = []
        other_chassis = []
        azs = set()

        # Check if candidates list valid
        if not candidates:
            return candidates

        chassis_with_azs = sb_idl.get_chassis_and_azs()

        # Get list of all AZs
        for chassis in candidates:
            try:
                azs.update(chassis_with_azs[chassis])
            except KeyError:
                continue

        for chassis in candidates:
            # Verify if chassis is in an AZ not already used
            # and delete AZs of chassis from list
            try:
                chassis_azs = chassis_with_azs[chassis]
                if azs.intersection(chassis_azs):
                    azs = azs.difference(chassis_azs)
                    chassis_selected += [chassis]
                else:
                    other_chassis += [chassis]
            except KeyError:
                other_chassis += [chassis]

        chassis_selected += other_chassis

        return chassis_selected

    @abc.abstractmethod
    def _select_gateway_chassis(self, nb_idl, sb_idl, candidates,
                                priority_min, priority_max, target_lrouter,
                                gateway_name=None):
        """Choose a chassis from candidates based on a specific policy.

        Returns a list of chassis to use for scheduling. The value at
        ``ret[0]`` will be used for the chassis with ``priority_max``, the
        value at ``ret[-1]`` will be used for the chassis with ``priority_min``
        """


class OVNGatewayChanceScheduler(OVNGatewayScheduler):
    """Randomly select an chassis for a gateway port of a router"""

    def select(self, nb_idl, sb_idl, gateway_name, candidates=None,
               existing_chassis=None, target_lrouter=None):
        return self._schedule_gateway(
            nb_idl, sb_idl, gateway_name,
            candidates, existing_chassis, target_lrouter)

    def _select_gateway_chassis(self, nb_idl, sb_idl, candidates,
                                priority_min, priority_max, target_lrouter,
                                gateway_name=None):
        candidates = copy.deepcopy(candidates)
        secrets.SystemRandom().shuffle(candidates)
        return self._reorder_by_az(nb_idl, sb_idl, candidates)


class OVNGatewayLeastLoadedScheduler(OVNGatewayScheduler):
    """Select the least loaded chassis for a gateway port of a router"""

    def select(self, nb_idl, sb_idl, gateway_name, candidates=None,
               existing_chassis=None, target_lrouter=None):
        return self._schedule_gateway(nb_idl, sb_idl, gateway_name,
                                      candidates, existing_chassis,
                                      target_lrouter)

    def _select_gateway_chassis(self, nb_idl, sb_idl, candidates,
                                priority_min, priority_max, target_lrouter,
                                gateway_name=None):
        """Returns a list of chassis from candidates ordered by priority
        (highest first). Each chassis in every priority will be selected, as it
        is the least loaded for that specific priority.
        """
        selected_chassis = []
        priorities = list(range(priority_max, priority_min - 1, -1))
        all_chassis_bindings = nb_idl.get_all_chassis_gateway_bindings(
                candidates, priorities=priorities)

        anti_affinity_score = 0
        chassis_hosting_lr = []

        lrouter_ports = getattr(target_lrouter, 'ports', set())
        if len(lrouter_ports):
            lrouter_ports_names = {getattr(lrp, 'name', "")
                                   for lrp in lrouter_ports}
            chassis_hosting_lr = [chassis
                for chassis, lrps in all_chassis_bindings.items() if
                not lrouter_ports_names.isdisjoint(
                    [lrp_name for lrp_name, prio in lrps])]

            # The `MAX_GW_CHASSIS` constant here is used mostly to get a
            # multiplier that guarantees our score will outweigh natural
            # LRP priority so that when other chassis are available those
            # will be chosen rather than a chassis already hosting a LRP
            # for this LR.
            anti_affinity_score = (ovn_const.MAX_GW_CHASSIS *
                                   len(target_lrouter.ports))

        for priority in priorities:
            chassis_load = {}
            for chassis, lrps in all_chassis_bindings.items():
                if chassis in selected_chassis:
                    continue
                lrps_with_prio = 0
                for lrp, prio in lrps:
                    if prio == priority:
                        lrps_with_prio += 1

                if chassis in chassis_hosting_lr:
                    chassis_load[chassis] = (lrps_with_prio +
                                             anti_affinity_score)
                else:
                    chassis_load[chassis] = lrps_with_prio
            if len(chassis_load) == 0:
                break

            leastload = min(chassis_load.values())
            least_loaded = [ch for ch, load in chassis_load.items()
                           if load == leastload]

            # Tie-break among chassis that share the minimum load at *this*
            # priority. Pure random (or hash of gateway name alone) is not
            # enough: greedy per-priority selection means prio-2 and prio-1
            # depend on what was chosen at higher priorities, so uncorrelated
            # tie-breaks can leave one chassis with too many bindings at a
            # single priority (see functional tests for multi-gateway routers).
            #
            # Strategy: sort so we prefer the chassis that is already heavier
            # at *lower* priorities (nearest lower priority first). That
            # pushes "overloaded below" chassis up to the current priority and
            # frees the lower slots for others, which self-corrects imbalance
            # across LRP scheduling rounds. When still tied (e.g. first cycle,
            # all zeros), use a deterministic hash of (gateway_name, priority)
            # XOR'd with hash(chassis) so each (LRP, priority) picks a stable
            # but distinct ordering among candidates.
            if len(least_loaded) > 1:
                lower_prios = [p for p in priorities if p < priority]

                if gateway_name:
                    key = '%s_%d' % (gateway_name, priority)
                    _hash = int(hashlib.sha512(
                        key.encode()).hexdigest(), 16)
                else:
                    _hash = secrets.randbits(128)

                def _sort_key(chassis, _lower_prios=lower_prios,
                              h=_hash):
                    loads = []
                    for lp in _lower_prios:
                        load_at_lp = sum(
                            1 for lrp_name, prio in
                            all_chassis_bindings.get(chassis, [])
                            if prio == lp)
                        # Descending sort on lower-priority counts: larger
                        # load_at_lp -> smaller tuple element -> sorts first.
                        loads.append(-load_at_lp)
                    # Final key: spread ties across chassis when counts match.
                    loads.append(h ^ hash(chassis))
                    return tuple(loads)

                least_loaded.sort(key=_sort_key)

            selected_chassis.append(least_loaded[0])

        return self._reorder_by_az(nb_idl, sb_idl, selected_chassis)


OVN_SCHEDULER_STR_TO_CLASS = {
    ovn_const.OVN_L3_SCHEDULER_CHANCE: OVNGatewayChanceScheduler,
    ovn_const.OVN_L3_SCHEDULER_LEASTLOADED: OVNGatewayLeastLoadedScheduler}


def get_scheduler():
    return OVN_SCHEDULER_STR_TO_CLASS[ovn_conf.get_ovn_l3_scheduler()]()



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import random

from oslo_db.sqlalchemy import utils as db_utils
from oslo_utils import uuidutils

from neutron.tests.functional.db import test_migrations


class NetworkSegmentRangesUniqueUpgrade(test_migrations.TestWalkMigrations):
    """Validates migration that adds unique constraints for
       network segment ranges.
    """

    _standard_attribute_id = 0

    def _gen_attr_id(self, type):
        self._standard_attribute_id = random.randint(100000, 2000000)
        standardattributes = db_utils.get_table(
            self.engine, 'standardattributes')
        with self.engine.connect() as conn, conn.begin():
            conn.execute(standardattributes.insert().values({
                'id': self._standard_attribute_id,
                'resource_type': type}))
        return self._standard_attribute_id

    def _create_network_segment_ranges(self, data):
        network_segment_ranges = db_utils.get_table(
            self.engine, 'network_segment_ranges')
        with self.engine.connect() as conn, conn.begin():
            for item in data:
                range_dict = {
                    'id': uuidutils.generate_uuid(),
                    'standard_attr_id': self._gen_attr_id(
                        'network_segment_ranges'),
                }
                range_dict.update(**item)
                conn.execute(
                    network_segment_ranges.insert().values(range_dict))

    def _pre_upgrade_b1bca967e19d(self, engine):
        duplicate_data = [
            {
                'name': '',
                'default': True,
                'shared': True,
                'network_type': 'vlan',
                'physical_network': 'default',
                'minimum': 100,
                'maximum': 200
            },
            {
                'name': '',
                'default': True,
                'shared': True,
                'network_type': 'vlan',
                'physical_network': 'default',
                'minimum': 100,
                'maximum': 200
            },
            {
                'name': '',
                'default': True,
                'shared': True,
                'network_type': 'vxlan',
                'minimum': 1000,
                'maximum': 2000
            },
            {
                'name': '',
                'default': True,
                'shared': True,
                'network_type': 'vxlan',
                'minimum': 1000,
                'maximum': 2000
            },
        ]
        self._create_network_segment_ranges(duplicate_data)
        # Ensure there are two duplicate ranges data
        range_table = db_utils.get_table(self.engine, 'network_segment_ranges')
        with self.engine.connect() as conn, conn.begin():
            rows = conn.execute(range_table.select()).fetchall()
            self.assertEqual(4, len(rows))
        return True

    def _check_b1bca967e19d(self, engine, data):
        range_table = db_utils.get_table(self.engine, 'network_segment_ranges')
        # check duplicate data is deleted
        with self.engine.connect() as conn, conn.begin():
            vlan = conn.execute(range_table.select().where(
                range_table.c.network_type == 'vlan')).fetchall()
            self.assertEqual(1, len(vlan))
            vxlan = conn.execute(range_table.select().where(
                range_table.c.network_type == 'vxlan')).fetchall()
            self.assertEqual(1, len(vxlan))



# Copyright 2025 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log
from ovsdbapp.backend.ovs_idl import connection
from ovsdbapp.backend.ovs_idl import idlutils
from ovsdbapp import event
from ovsdbapp.schema.ovn_northbound import impl_idl as nb_impl_idl
from ovsdbapp.schema.ovn_southbound import impl_idl as sb_impl_idl

from neutron.plugins.ml2.drivers.ovn.mech_driver.ovsdb import ovsdb_monitor

LOG = log.getLogger(__name__)

OVN_NB_TABLES = (
    'Logical_Switch', 'Logical_Switch_Port',
    'Logical_Router', 'Logical_Router_Port',
    'HA_Chassis_Group', 'HA_Chassis',
    'Logical_Router_Static_Route', 'Logical_Router_Policy',
    'DHCP_Options',
)
OVN_SB_TABLES = ('Chassis', 'Chassis_Private')


class OvnIdl(connection.OvsdbIdl):
    LEADER_ONLY = False

    def __init__(self, connection_string, events=None):
        if connection_string.startswith("ssl"):
            ovsdb_monitor._check_and_set_ssl_files(self.SCHEMA)
        helper = idlutils.get_schema_helper(connection_string, self.SCHEMA)
        for table in self.tables:
            helper.register_table(table)
        self.notify_handler = event.RowEventHandler()
        if events:
            self.notify_handler.watch_events(events)
        super().__init__(
            connection_string, helper, leader_only=self.LEADER_ONLY)

    def notify(self, event, row, updates=None):
        self.notify_handler.notify(event, row, updates)

    def start(self, timeout):
        conn = connection.Connection(self, timeout=timeout)
        return self.api_cls(conn)


class BgpOvnNbIdl(nb_impl_idl.OvnNbApiIdlImpl):
    @property
    def has_lock(self):
        return self.ovsdb_connection.idl.has_lock

    def register_events(self, events):
        self.ovsdb_connection.idl.notify_handler.watch_events(events)


class BgpOvnSbIdl(sb_impl_idl.OvnSbApiIdlImpl):
    def stop(self):
        self.ovsdb_connection.stop()

    def register_events(self, events):
        self.ovsdb_connection.idl.notify_handler.watch_events(events)


class OvnNbIdl(OvnIdl):
    LOCK_NAME = 'bgp_topology_lock'
    LEADER_ONLY = True
    SCHEMA = 'OVN_Northbound'
    tables = OVN_NB_TABLES
    api_cls = BgpOvnNbIdl

    def start(self, timeout):
        LOG.debug("Setting lock for BGP topology")
        self.set_lock(self.LOCK_NAME)
        return super().start(timeout)


class OvnSbIdl(OvnIdl):
    SCHEMA = 'OVN_Southbound'
    tables = OVN_SB_TABLES
    api_cls = BgpOvnSbIdl



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import ddt
from neutron_lib.api.definitions import uplink_status_propagation as usp
from neutron_lib.api.definitions import uplink_status_propagation_updatable \
    as apidef
from neutron_lib.db import api as db_api
from neutron_lib.db import resource_extend

from neutron.db import db_base_plugin_v2
from neutron.db import uplink_status_propagation_db as usp_db
from neutron.tests.common import test_db_base_plugin_v2


class UplinkStatusPropagationUpdatableExtensionTestPlugin(
        db_base_plugin_v2.NeutronDbPluginV2,
        usp_db.UplinkStatusPropagationMixin):
    """Test plugin to mixin the uplink status propagation extension.
    """

    supported_extension_aliases = [usp.ALIAS, apidef.ALIAS]

    # TODO(ralonsoh): update ``uplink_status_propagation_updatable`` with
    # COLLECTION_NAME=neutron_lib.api.definitions.port.COLLECTION_NAME
    @staticmethod
    @resource_extend.extends([usp.COLLECTION_NAME])
    def _extend_network_project_default(port_res, port_db):
        return usp_db.UplinkStatusPropagationMixin._extend_port_dict(
            port_res, port_db)

    def create_port(self, context, port):
        with db_api.CONTEXT_WRITER.using(context):
            new_port = super().create_port(context, port)
            # Update the propagate_uplink_status in the database
            p = port['port']
            if 'propagate_uplink_status' not in p:
                p['propagate_uplink_status'] = False
            self._process_create_port(context, p, new_port)
        return new_port

    def update_port(self, context, port_id, port, **kwargs):
        with db_api.CONTEXT_WRITER.using(context):
            new_port = super().update_port(context, port_id, port)
            # Update the propagate_uplink_status in the database
            p = port['port']
            if 'propagate_uplink_status' not in p:
                p['propagate_uplink_status'] = False
            self._process_update_port(context, p, new_port)
        return new_port


@ddt.ddt
class UplinkStatusPropagationUpdatableExtensionTestCase(
         test_db_base_plugin_v2.NeutronDbPluginV2TestCase):
    """Test API extension ``uplink-status-propagation-updatable`` attributes.
    """

    def setUp(self, **kwargs):
        plugin = ('neutron.tests.unit.extensions.test_uplink_status_'
                  'propagation_updatable.'
                  'UplinkStatusPropagationUpdatableExtensionTestPlugin')
        super().setUp(plugin=plugin)

    @ddt.data(True, False)
    def test_update_port_propagate_uplink_status(
            self, propagate_uplink_status):
        name = 'propagate_uplink_status'
        keys = [('name', name), ('admin_state_up', True),
                ('status', self.port_create_status),
                (usp.PROPAGATE_UPLINK_STATUS, propagate_uplink_status)]
        with self.port(name=name,
                       propagate_uplink_status=propagate_uplink_status
                       ) as port:
            for k, v in keys:
                self.assertEqual(v, port['port'][k])

            # Update the port with the opposite ``propagate_uplink_status``
            # value and check it.
            data = {'port': {usp.PROPAGATE_UPLINK_STATUS:
                             not propagate_uplink_status}}
            req = self.new_update_request('ports', data, port['port']['id'])
            req.get_response(self.api)
            req = self.new_show_request('ports', port['port']['id'])
            res = req.get_response(self.api)
            port = self.deserialize(self.fmt, res)['port']
            self.assertEqual(not propagate_uplink_status,
                             port[usp.PROPAGATE_UPLINK_STATUS])



# Copyright 2012 New Dream Network, LLC (DreamHost)
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc
from concurrent import futures
import socketserver
import urllib

import netaddr
from neutron_lib import constants
from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import netutils
import requests
import webob

from neutron._i18n import _
from neutron.agent.linux import utils as agent_utils
from neutron.common import ipv6_utils
from neutron.common import utils as common_utils
from neutron.conf.agent.metadata import config

LOG = logging.getLogger(__name__)

MODE_MAP = {
    config.USER_MODE: 0o644,
    config.GROUP_MODE: 0o664,
    config.ALL_MODE: 0o666,
}


class MetadataProxyServer(socketserver.UnixStreamServer):
    """Metadata server which listens on a unix domain socket.

    This is based on the ThreadingMixIn, but is capable to
      - process requests within the main process when workers <= 0
      - process requests using a threadpool when workers > 0
    """

    def __init__(self, workers, *kargs, **kwargs):
        super().__init__(*kargs, **kwargs)
        self._pool = None
        if workers > 0:
            self._pool = futures.ThreadPoolExecutor(max_workers=workers)
        self.request_queue_size = cfg.CONF.metadata_backlog

    def process_request_thread(self, request, client_address):
        """Same as in BaseServer but as a thread.

        In addition, exception handling is done here.

        """
        try:
            self.finish_request(request, client_address)
        except Exception:
            self.handle_error(request, client_address)
        finally:
            self.shutdown_request(request)

    def process_request(self, request, client_address):
        """Start a new thread to process the request."""
        if self._pool is not None:
            self._pool.submit(self.process_request_thread, request,
                              client_address)
        else:
            # NOTE(tkajinam): Run within the proces when workers<1
            super().process_request(request, client_address)

    def server_close(self):
        super().server_close()
        if self._pool is not None:
            self._pool.shutdown()


class MetadataProxyHandlerBase(metaclass=abc.ABCMeta):
    NETWORK_ID_HEADER: str
    ROUTER_ID_HEADER: str

    def __init__(self, conf, has_cache=False, **kwargs):
        self.conf = conf
        self._has_cache = has_cache
        super().__init__(**kwargs)

    @abc.abstractmethod
    def get_port(self, remote_address, network_id=None, remote_mac=None,
                 router_id=None, skip_cache=False):
        """Search for a single port that contain the given IP address and
        belongs to the given network.

        If no network is passed, ports are searched on all networks connected
        to a given router. Either one of network_id or router_id must be given.

        :param remote_address: IP address to search for
        :param network_id: Network ID to filter by, if given
        :param remote_mac: Remote MAC to filter by, if given
        :param router_id: Router ID to filter by, if given
        :param skip_cache: When to skip getting entry from cache

        """
        pass

    @webob.dec.wsgify(RequestClass=webob.Request)
    def __call__(self, req):
        try:
            LOG.debug("Request: %s", req)

            instance_id, project_id = self._get_instance_and_project_id(req)
            if instance_id:
                res = self._proxy_request(instance_id, project_id, req)
                if isinstance(res, webob.exc.HTTPNotFound) and self._has_cache:
                    LOG.info("The instance: %s is not present anymore, "
                             "skipping cache...", instance_id)
                    instance_id, project_id = (
                        self._get_instance_and_project_id(req,
                                                          skip_cache=True))
                    if instance_id:
                        res = self._proxy_request(instance_id, project_id, req)
                return res
            return webob.exc.HTTPNotFound()

        except Exception:
            LOG.exception("Unexpected error.")
            msg = _('An unknown error has occurred. '
                    'Please try your request again.')
            explanation = str(msg)
            return webob.exc.HTTPInternalServerError(explanation=explanation)

    def _get_instance_id(self, req):
        """Returns the network ID and the router ID from the request"""
        network_id = req.headers.get(self.NETWORK_ID_HEADER)
        router_id = (req.headers.get(self.ROUTER_ID_HEADER)
                     if self.ROUTER_ID_HEADER else None)
        return network_id, router_id

    def _get_instance_and_project_id(self, req, skip_cache=False):
        forwarded_for = req.headers.get('X-Forwarded-For')
        network_id, router_id = self._get_instance_id(req)

        # Only one should be given, drop since it could be spoofed
        if network_id and router_id:
            LOG.debug("Both network and router IDs were specified in proxy "
                      "request, but only a single one of the two is allowed, "
                      "dropping")
            return None, None

        remote_mac = None
        remote_ip = netaddr.IPAddress(forwarded_for)
        if remote_ip.version == constants.IP_VERSION_6:
            if remote_ip.is_ipv4_mapped():
                # When haproxy listens on v4 AND v6 then it inserts ipv4
                # addresses as ipv4-mapped v6 addresses into X-Forwarded-For.
                forwarded_for = str(remote_ip.ipv4())
            if remote_ip.is_link_local():
                # When haproxy sees an ipv6 link-local client address
                # (and sends that to us in X-Forwarded-For) we must rely
                # on the EUI encoded in it, because that's all we can
                # recognize.
                remote_mac = str(netutils.get_mac_addr_by_ipv6(remote_ip))

        instance_id, project_id = self.get_port(forwarded_for,
                                                network_id=network_id,
                                                remote_mac=remote_mac,
                                                router_id=router_id,
                                                skip_cache=skip_cache)
        return instance_id, project_id

    def _proxy_request(self, instance_id, project_id, req):
        headers = {
            'X-Forwarded-For': req.headers.get('X-Forwarded-For'),
            'X-Instance-ID': instance_id,
            'X-Tenant-ID': project_id,
            'X-Instance-ID-Signature': common_utils.sign_instance_id(
                self.conf, instance_id)
        }

        nova_host_port = ipv6_utils.valid_ipv6_url(
            self.conf.nova_metadata_host,
            self.conf.nova_metadata_port)

        url = urllib.parse.urlunsplit((
            self.conf.nova_metadata_protocol,
            nova_host_port,
            req.path_info,
            req.query_string,
            ''))

        disable_ssl_certificate_validation = self.conf.nova_metadata_insecure
        if self.conf.auth_ca_cert and not disable_ssl_certificate_validation:
            verify_cert = self.conf.auth_ca_cert
        else:
            verify_cert = not disable_ssl_certificate_validation

        client_cert = None
        if self.conf.nova_client_cert and self.conf.nova_client_priv_key:
            client_cert = (self.conf.nova_client_cert,
                           self.conf.nova_client_priv_key)

        try:
            resp = requests.request(method=req.method, url=url,
                                    headers=headers,
                                    data=req.body,
                                    cert=client_cert,
                                    verify=verify_cert,
                                    timeout=60)
        except requests.ConnectionError:
            msg = _('The remote metadata server is temporarily unavailable. '
                    'Please try again later.')
            explanation = str(msg)
            return webob.exc.HTTPServiceUnavailable(explanation=explanation)

        if resp.status_code == 200:
            req.response.content_type = resp.headers['content-type']
            req.response.body = resp.content
            LOG.debug(str(resp))
            return req.response
        if resp.status_code == 403:
            LOG.warning(
                'The remote metadata server responded with Forbidden. This '
                'response usually occurs when shared secrets do not match.'
            )
            return webob.exc.HTTPForbidden()
        if resp.status_code == 500:
            msg = _(
                'Remote metadata server experienced an internal server error.'
            )
            LOG.warning(msg)
            explanation = str(msg)
            return webob.exc.HTTPInternalServerError(explanation=explanation)
        if resp.status_code in (400, 404, 409, 502, 503, 504):
            webob_exc_cls = webob.exc.status_map.get(resp.status_code)
            return webob_exc_cls()
        raise Exception(_('Unexpected response code: %s') % resp.status_code)


class UnixDomainMetadataProxyBase(metaclass=abc.ABCMeta):

    def __init__(self, conf):
        self.conf = conf

    def _get_socket_mode(self):
        mode = self.conf.metadata_proxy_socket_mode
        if mode == config.DEDUCE_MODE:
            user = self.conf.metadata_proxy_user
            if (not user or user == '0' or user == 'root' or
                    agent_utils.is_effective_user(user)):
                # user is agent effective user or root => USER_MODE
                mode = config.USER_MODE
            else:
                group = self.conf.metadata_proxy_group
                if not group or agent_utils.is_effective_group(group):
                    # group is agent effective group => GROUP_MODE
                    mode = config.GROUP_MODE
                else:
                    # otherwise => ALL_MODE
                    mode = config.ALL_MODE
        return MODE_MAP[mode]

    @abc.abstractmethod
    def run(self):
        pass



# Copyright 2026 Red Hat, LLC
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from neutron.agent.ovn.extensions import bgp as bgp_ext
from neutron.tests import base


class TestBGPAgentExtensionWatchPortCreatedEvent(base.BaseTestCase):
    def setUp(self):
        super().setUp()
        self.extension = bgp_ext.BGPAgentExtension()
        self.extension.agent_api = mock.Mock()
        self.event_handler = (
            self.extension.agent_api.ovs_idl.idl.notify_handler)

        self.bgp_bridge = mock.Mock()
        self.bgp_bridge.name = 'br-bgp'

    def test_port_already_exists_configures_flows(self):
        self.bgp_bridge.ovs_bridge.get_iface_ofports_by_type.return_value = [5]
        self.bgp_bridge.check_requirements_for_flows_met.return_value = True

        self.extension.watch_port_created_event(self.bgp_bridge, 'patch')

        self.bgp_bridge.configure_flows.assert_called_once()

    def test_port_already_exists_requirements_not_met_no_configure_flows(self):
        self.bgp_bridge.ovs_bridge.get_iface_ofports_by_type.return_value = [5]
        self.bgp_bridge.check_requirements_for_flows_met.return_value = False

        self.extension.watch_port_created_event(self.bgp_bridge, 'patch')

        self.bgp_bridge.configure_flows.assert_not_called()

    def test_port_missing_watches_event(self):
        self.bgp_bridge.ovs_bridge.get_iface_ofports_by_type.return_value = []

        self.extension.watch_port_created_event(self.bgp_bridge, 'patch')

        self.event_handler.watch_event.assert_called_once()
        self.bgp_bridge.configure_flows.assert_not_called()

    def test_port_created_in_meantime_unwatches_and_configures_flows(self):
        self.bgp_bridge.ovs_bridge.get_iface_ofports_by_type.side_effect = [
            [],
            [7],
        ]
        self.bgp_bridge.check_requirements_for_flows_met.return_value = True

        self.extension.watch_port_created_event(self.bgp_bridge, 'patch')

        self.event_handler.watch_event.assert_called_once()
        self.event_handler.unwatch_event.assert_called_once()
        self.bgp_bridge.configure_flows.assert_called_once()

    def test_port_created_in_meantime_requirements_not_met_no_configure_flows(
            self):
        self.bgp_bridge.ovs_bridge.get_iface_ofports_by_type.side_effect = [
            [],
            [3],
        ]
        self.bgp_bridge.check_requirements_for_flows_met.return_value = False

        self.extension.watch_port_created_event(self.bgp_bridge, 'patch')

        self.event_handler.watch_event.assert_called_once()
        self.event_handler.unwatch_event.assert_called_once()
        self.bgp_bridge.configure_flows.assert_not_called()



# Copyright 2025 Red Hat, LLC
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from ovsdbapp.backend.ovs_idl import event


class WaitForPortBindingEvent(event.WaitEvent):
    event_name = 'WaitForPortBindingEvent'

    def __init__(self, port_name):
        table = 'Port_Binding'
        events = (self.ROW_CREATE,)
        conditions = (('logical_port', '=', port_name),)
        super().__init__(events, table, conditions, timeout=10)



# Copyright (c) 2013 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from neutron_lib.api import attributes
from neutron_lib.api.definitions import external_net as extnet_apidef
from neutron_lib.api.definitions import subnet as subnet_apidef
from neutron_lib.api.definitions import subnet_external_network as \
    extsnet_apidef
from neutron_lib import constants
from neutron_lib import context
from neutron_lib.plugins import constants as plugin_constants
from neutron_lib.plugins import directory
from oslo_utils import uuidutils
import testtools
from webob import exc

from neutron.tests.unit.api.v2 import test_base


# Add subnet 'router:external' extension, without loading the extensions.
# This change must be done before the policies are parsed in order to load the
# 'convert_to' method before the ``FieldCheck`` instance for this field is
# created.
rname = subnet_apidef.COLLECTION_NAME
attributes.RESOURCES[rname].update(
    extsnet_apidef.RESOURCE_ATTRIBUTE_MAP[rname])
from neutron.tests.common import test_db_base_plugin_v2  # noqa: E402


_uuid = uuidutils.generate_uuid
_get_path = test_base._get_path


class ExtNetTestExtensionManager:

    def get_resources(self):
        return []

    def get_actions(self):
        return []

    def get_request_extensions(self):
        return []


class ExtNetDBTestCase(test_db_base_plugin_v2.NeutronDbPluginV2TestCase):

    def _create_network(self, fmt, name, admin_state_up, **kwargs):
        """Override the routine for allowing the router:external attribute."""
        # attributes containing a colon should be passed with
        # a double underscore
        new_args = dict(zip(map(lambda x: x.replace('__', ':'), kwargs),
                            kwargs.values()))
        arg_list = new_args.pop('arg_list', ()) + (extnet_apidef.EXTERNAL,)
        return super()._create_network(
            fmt, name, admin_state_up, arg_list=arg_list, **new_args)

    def setUp(self):
        plugin = 'neutron.tests.unit.extensions.test_l3.TestNoL3NatPlugin'
        ext_mgr = ExtNetTestExtensionManager()
        super().setUp(plugin=plugin, ext_mgr=ext_mgr)

    def _set_net_external(self, net_id):
        self._update('networks', net_id,
                     {'network': {extnet_apidef.EXTERNAL: True}},
                     as_admin=True)

    def test_list_nets_external(self):
        with self.network() as n1:
            self._set_net_external(n1['network']['id'])
            with self.network():
                body = self._list('networks')
                self.assertEqual(2, len(body['networks']))

                body = self._list('networks',
                                  query_params="%s=True" %
                                               extnet_apidef.EXTERNAL)
                self.assertEqual(1, len(body['networks']))

                body = self._list('networks',
                                  query_params="%s=False" %
                                               extnet_apidef.EXTERNAL)
                self.assertEqual(1, len(body['networks']))

    def test_list_nets_external_pagination(self):
        if self._skip_native_pagination:
            self.skipTest("Skip test for not implemented pagination feature")
        with self.network(name='net1') as n1, self.network(name='net3') as n3:
            self._set_net_external(n1['network']['id'])
            self._set_net_external(n3['network']['id'])
            with self.network(name='net2') as n2:
                self._test_list_with_pagination(
                    'network', (n1, n3), ('name', 'asc'), 1, 3,
                    query_params='router:external=True')
                self._test_list_with_pagination(
                    'network', (n2, ), ('name', 'asc'), 1, 2,
                    query_params='router:external=False')

    def test_get_network_succeeds_without_filter(self):
        plugin = directory.get_plugin()
        ctx = context.Context(None, None, is_admin=True)
        result = plugin.get_networks(ctx, filters=None)
        self.assertEqual([], result)

    def test_update_network_set_external_non_admin_fails(self):
        # Assert that a non-admin user cannot update the
        # router:external attribute
        with self.network(project_id='noadmin') as network:
            data = {'network': {'router:external': True}}
            req = self.new_update_request('networks',
                                          data,
                                          network['network']['id'],
                                          project_id='noadmin')
            res = req.get_response(self.api)
            self.assertEqual(exc.HTTPForbidden.code, res.status_int)

    def test_update_network_external_net_with_ports_set_not_shared(self):
        with self.network(router__external=True, shared=True,
                          as_admin=True) as ext_net,\
                self.subnet(network=ext_net) as ext_subnet, \
                self.port(subnet=ext_subnet,
                          project_id='',
                          device_owner=constants.DEVICE_OWNER_ROUTER_SNAT):
            data = {'network': {'shared': False}}
            req = self.new_update_request('networks',
                                          data,
                                          ext_net['network']['id'],
                                          as_admin=True)
            res = req.get_response(self.api)
            self.assertEqual(exc.HTTPOk.code, res.status_int)
            ctx = context.Context(None, None, is_admin=True)
            plugin = directory.get_plugin()
            result = plugin.get_networks(ctx)
            self.assertFalse(result[0]['shared'])

    def test_create_port_external_network_non_admin_fails(self):
        with self.network(as_admin=True, router__external=True) as ext_net:
            with self.subnet(network=ext_net) as ext_subnet:
                with testtools.ExpectedException(
                        exc.HTTPClientError) as ctx_manager:
                    with self.port(subnet=ext_subnet,
                                   is_admin=False,
                                   project_id='noadmin'):
                        pass
                    self.assertEqual(403, ctx_manager.exception.code)

    def test_create_port_external_network_admin_succeeds(self):
        with self.network(router__external=True, as_admin=True) as ext_net:
            with self.subnet(network=ext_net) as ext_subnet:
                with self.port(subnet=ext_subnet) as port:
                    self.assertEqual(port['port']['network_id'],
                                     ext_net['network']['id'])

    def test_create_external_network_non_admin_fails(self):
        with testtools.ExpectedException(exc.HTTPClientError) as ctx_manager:
            with self.network(router__external=True,
                              as_admin=False,
                              project_id='noadmin'):
                pass
            self.assertEqual(403, ctx_manager.exception.code)

    def test_create_external_network_admin_succeeds(self):
        with self.network(router__external=True, as_admin=True) as ext_net:
            self.assertTrue(ext_net['network'][extnet_apidef.EXTERNAL])

    def test_delete_network_check_disassociated_floatingips(self):
        l3_mock = mock.Mock()
        directory.add_plugin(plugin_constants.L3, l3_mock)
        with self.network() as net:
            req = self.new_delete_request('networks', net['network']['id'])
            res = req.get_response(self.api)
            self.assertEqual(exc.HTTPNoContent.code, res.status_int)
            (l3_mock.delete_disassociated_floatingips
             .assert_called_once_with(mock.ANY, net['network']['id']))

    def test_create_shared_networks_and_subnets(self):
        with (self.network(as_admin=True, router__external=True) as net_ext,
                self.network(as_admin=True, shared=True) as net_shared,
                self.network(as_admin=True) as net_admin):
            with (self.subnet(as_admin=True, network=net_ext)
                  as snet_ext, self.subnet(as_admin=True, network=net_shared)
                  as snet_shared, self.subnet(as_admin=True, network=net_admin)
                  as snet_admin):
                req = self.new_list_request('networks', as_admin=False,
                                            project_id='noadmin')
                res = self.deserialize(self.fmt, req.get_response(self.api))
                net_ids = {net['id'] for net in res['networks']}
                self.assertIn(net_ext['network']['id'], net_ids)
                self.assertIn(net_shared['network']['id'], net_ids)
                self.assertNotIn(net_admin['network']['id'], net_ids)

                req = self.new_list_request('subnets', as_admin=False,
                                            project_id='noadmin')
                res = self.deserialize(self.fmt, req.get_response(self.api))
                snet_ids = {snet['id'] for snet in res['subnets']}
                self.assertIn(snet_ext['subnet']['id'], snet_ids)
                self.assertIn(snet_shared['subnet']['id'], snet_ids)
                self.assertNotIn(snet_admin['subnet']['id'], snet_ids)



# Copyright 2025 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron.services.bgp import reconciler
from neutron import worker


class BGPWorker(worker.NeutronBaseWorker):
    def __init__(self):
        super().__init__(worker_process_count=0)

    def start(self):
        super().start(desc="bgp worker")
        self._reconciler = reconciler.BGPTopologyReconciler()

    def wait(self):
        self._reconciler.full_sync()

    def stop(self):
        self._reconciler.stop()

    def reset(self):
        pass



# Copyright 2014 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections
from contextlib import contextmanager
import subprocess

from alembic.ddl import base as alembic_ddl
from alembic import script as alembic_script
from oslo_config import cfg
from oslo_config import fixture as config_fixture
from oslo_db.sqlalchemy import test_migrations
from oslo_log import log as logging
from oslotest import base as oslotest_base
import sqlalchemy
from sqlalchemy import event  # noqa
from sqlalchemy.sql import ddl as sqla_ddl

from neutron.common import config
from neutron.db import migration as migration_root
from neutron.db.migration.alembic_migrations import external
from neutron.db.migration import cli as migration
from neutron.db.migration.models import head as head_models
from neutron.tests import base as test_base
from neutron.tests.functional import base as functional_base
from neutron.tests.unit import testlib_api


CREATION_OPERATIONS = {
    'sqla': (sqla_ddl.CreateIndex,
             sqla_ddl.CreateTable,
             sqla_ddl.CreateColumn,
             ),
    'alembic': (alembic_ddl.AddColumn,
                )
}

DROP_OPERATIONS = {
    'sqla': (sqla_ddl.DropConstraint,
             sqla_ddl.DropIndex,
             sqla_ddl.DropTable,
             ),
    'alembic': (alembic_ddl.DropColumn,
                )
}

LOG = logging.getLogger(__name__)

# NOTE(slaweq): replace alembic_util logging functions used normally with
# olso_log logger to limit output on stdout
migration.log_error = LOG.error
migration.log_warning = LOG.warning
migration.log_info = LOG.info


def render_url_str(_url):
    """Render a ``URL`` instance as string and the password in clear text"""
    try:
        return _url.render_as_string(hide_password=False)
    except AttributeError:
        # NOTE(ralonsoh): ``URL`` objects from SQLAlchemy<2.0.0 don't have
        # ``render_as_string`` method but it is not necessary.
        return str(_url)


def upgrade(engine, alembic_config, branch_name='heads'):
    url_str = render_url_str(engine.url)
    cfg.CONF.set_override('connection', url_str, group='database')
    migration.do_alembic_command(alembic_config, 'upgrade',
                                 branch_name)


class TestModelsMigrations(test_migrations.ModelsMigrationsSync,
                           testlib_api.MySQLTestCaseMixin,
                           testlib_api.SqlTestCaseLight,
                           functional_base.BaseLoggingTestCase):
    '''Test for checking of equality models state and migrations.

    For the opportunistic testing you need to set up a db named
    'openstack_citest' with user 'openstack_citest' and password
    'openstack_citest' on localhost.
    The test will then use that db and user/password combo to run the tests.

    For MySQL on Ubuntu this can be done with the following commands::

        mysql -u root
        >create database openstack_citest;
        >grant all privileges on openstack_citest.* to
         openstack_citest@localhost identified by 'openstack_citest';

    Output is a list that contains information about differences between db and
    models. Output example::

       [('add_table',
         Table('bat', MetaData(bind=None),
               Column('info', String(), table=<bat>), schema=None)),
        ('remove_table',
         Table(u'bar', MetaData(bind=None),
               Column(u'data', VARCHAR(), table=<bar>), schema=None)),
        ('add_column',
         None,
         'foo',
         Column('data', Integer(), table=<foo>)),
        ('remove_column',
         None,
         'foo',
         Column(u'old_data', VARCHAR(), table=None)),
        [('modify_nullable',
          None,
          'foo',
          u'x',
          {'existing_server_default': None,
          'existing_type': INTEGER()},
          True,
          False)]]

    * ``remove_*`` means that there is extra table/column/constraint in db;

    * ``add_*`` means that it is missing in db;

    * ``modify_*`` means that on column in db is set wrong
      type/nullable/server_default. Element contains information:

        - what should be modified,
        - schema,
        - table,
        - column,
        - existing correct column parameters,
        - right value,
        - wrong value.

    This class also contains tests for branches, like that correct operations
    are used in contract and expand branches.

    '''

    BUILD_SCHEMA = False
    TIMEOUT_SCALING_FACTOR = 4

    def setUp(self):
        config.register_common_config_options()
        super().setUp()
        self.cfg = self.useFixture(config_fixture.Config())
        self.cfg.config(core_plugin='ml2')
        self.alembic_config = migration.get_neutron_config()
        self.alembic_config.neutron_config = cfg.CONF

        # Migration tests can take a long time
        self.useFixture(test_base.Timeout(scaling=self.TIMEOUT_SCALING_FACTOR))

    def db_sync(self, engine):
        upgrade(engine, self.alembic_config)

    def get_engine(self):
        return self.engine

    def get_metadata(self):
        return head_models.get_metadata()

    def include_object(self, object_, name, type_, reflected, compare_to):
        if type_ == 'table' and (name == 'alembic_version' or
                                 name in external.TABLES):
            return False

        return super().include_object(
            object_, name, type_, reflected, compare_to)

    def filter_metadata_diff(self, diff):
        return list(filter(self.remove_unrelated_errors, diff))

    # Remove some difference that are not mistakes just specific of
    # dialects, etc
    def remove_unrelated_errors(self, element):
        insp = sqlalchemy.inspect(self.get_engine())
        dialect = self.get_engine().dialect.name
        if isinstance(element, tuple):
            if dialect == 'mysql' and element[0] == 'remove_index':
                table_name = element[1].table.name
                for fk in insp.get_foreign_keys(table_name):
                    if fk['name'] == element[1].name:
                        return False
                cols = [c.name for c in element[1].expressions]
                for col in cols:
                    if col in insp.get_pk_constraint(
                            table_name)['constrained_columns']:
                        return False
        else:
            for modified, _, table, column, _, _, new in element:
                if modified == 'modify_default' and dialect == 'mysql':
                    constrained = insp.get_pk_constraint(table)
                    if column in constrained['constrained_columns']:
                        return False
        return True

    def test_upgrade_expand_branch(self):
        # Verify that "command neutron-db-manage upgrade --expand" works
        # without errors.
        upgrade(self.engine, self.alembic_config,
                branch_name='%s@head' % migration.EXPAND_BRANCH)

    def test_upgrade_contract_branch(self):
        # Verify that "command neutron-db-manage upgrade --contract" works
        # without errors.
        upgrade(self.engine, self.alembic_config,
                branch_name='%s@head' % migration.CONTRACT_BRANCH)

    @contextmanager
    def _listener(self, engine, listener_func):
        try:
            event.listen(engine, 'before_execute', listener_func)
            yield
        finally:
            event.remove(engine, 'before_execute',
                         listener_func)

    def test_branches(self):

        drop_exceptions = collections.defaultdict(list)
        creation_exceptions = collections.defaultdict(list)

        def find_migration_exceptions():
            # Due to some misunderstandings and some conscious decisions,
            # there may be some expand migrations which drop elements and
            # some contract migrations which create elements. These excepted
            # elements must be returned by a method in the script itself.
            # The names of the method must be 'contract_creation_exceptions'
            # or 'expand_drop_exceptions'. The methods must have a docstring
            # explaining the reason for the exception.
            #
            # Here we build lists of the excepted elements and verify that
            # they are documented.
            script = alembic_script.ScriptDirectory.from_config(
                self.alembic_config)
            for m in list(script.walk_revisions(base='base', head='heads')):
                branches = m.branch_labels or []
                if migration.CONTRACT_BRANCH in branches:
                    method_name = 'contract_creation_exceptions'
                    exceptions_dict = creation_exceptions
                elif migration.EXPAND_BRANCH in branches:
                    method_name = 'expand_drop_exceptions'
                    exceptions_dict = drop_exceptions
                else:
                    continue
                get_excepted_elements = getattr(m.module, method_name, None)
                if not get_excepted_elements:
                    continue
                explanation = getattr(get_excepted_elements, '__doc__', "")
                if not explanation:
                    self.fail("%s() requires docstring with explanation" %
                              '.'.join([m.module.__name__,
                                        get_excepted_elements.__name__]))
                for sa_type, elements in get_excepted_elements().items():
                    exceptions_dict[sa_type].extend(elements)

        def is_excepted_sqla(clauseelement, exceptions):
            """Identify excepted operations that are allowed for the branch."""
            element = clauseelement.element
            element_name = element.name
            if isinstance(element, sqlalchemy.Index):
                element_name = element.table.name
            for sa_type_, excepted_names in exceptions.items():
                if isinstance(element, sa_type_):
                    if element_name in excepted_names:
                        return True

        def is_excepted_alembic(clauseelement, exceptions):
            """Identify excepted operations that are allowed for the branch."""
            # For alembic the clause is AddColumn or DropColumn
            column = clauseelement.column.name
            table = clauseelement.table_name
            element_name = '.'.join([table, column])
            for alembic_type, excepted_names in exceptions.items():
                if alembic_type == sqlalchemy.Column:
                    if element_name in excepted_names:
                        return True

        def is_allowed(clauseelement, exceptions, disallowed_ops):
            if (isinstance(clauseelement, disallowed_ops['sqla']) and
                    hasattr(clauseelement, 'element')):
                return is_excepted_sqla(clauseelement, exceptions)
            if isinstance(clauseelement, disallowed_ops['alembic']):
                return is_excepted_alembic(clauseelement, exceptions)
            return True

        def check_expand_branch(conn, clauseelement, multiparams, params,
                                execution_options):
            if not is_allowed(clauseelement, drop_exceptions, DROP_OPERATIONS):
                self.fail("Migration in expand branch contains drop command")

        def check_contract_branch(conn, clauseelement, multiparams, params,
                                  execution_options):
            if not is_allowed(clauseelement, creation_exceptions,
                              CREATION_OPERATIONS):
                self.fail("Migration in contract branch contains create "
                          "command")

        find_migration_exceptions()
        engine = self.engine
        url_str = render_url_str(engine.url)
        cfg.CONF.set_override('connection', url_str, group='database')

        with engine.begin() as connection:
            self.alembic_config.attributes['connection'] = connection

            # upgrade to latest release first; --expand users are expected to
            # apply all alembic scripts from previous releases before applying
            # the new ones
            for release in migration_root.NEUTRON_MILESTONES:
                release_revisions = migration._find_milestone_revisions(
                    self.alembic_config, release)
                for rev in release_revisions:
                    migration.do_alembic_command(
                        self.alembic_config, 'upgrade', rev[0])

            with self._listener(engine, check_expand_branch):
                migration.do_alembic_command(
                    self.alembic_config, 'upgrade',
                    '%s@head' % migration.EXPAND_BRANCH)

            with self._listener(engine, check_contract_branch):
                migration.do_alembic_command(
                    self.alembic_config, 'upgrade',
                    '%s@head' % migration.CONTRACT_BRANCH)

    # NOTE(ihrachys): if this test fails for you, it probably means that you
    # attempt to add an unsafe contract migration script, that is in
    # contradiction to blueprint online-upgrades
    def test_forbid_offline_migrations_starting_newton(self):
        engine = self.get_engine()
        url_str = render_url_str(engine.url)
        cfg.CONF.set_override('connection', url_str, group='database')
        # the following revisions are Newton heads
        for revision in ('5cd92597d11d', '5c85685d616d'):
            migration.do_alembic_command(
                self.alembic_config, 'upgrade', revision)
        self.assertFalse(migration.has_offline_migrations(
            self.alembic_config, 'unused'),
            msg='Offline contract migration scripts are forbidden for Ocata+')

    def test_check_mysql_engine(self):
        engine = self.get_engine()
        url_str = render_url_str(engine.url)
        cfg.CONF.set_override('connection', url_str, group='database')
        with engine.begin() as connection:
            self.alembic_config.attributes['connection'] = connection
            migration.do_alembic_command(self.alembic_config, 'upgrade',
                                         'heads')
            insp = sqlalchemy.inspect(engine)
            # Test that table creation on MySQL only builds InnoDB tables
            tables = insp.get_table_names()
            self.assertGreater(len(tables), 0,
                               "No tables found. Wrong schema?")
            res = [table for table in tables if
                   insp.get_table_options(table)['mysql_engine'] !=
                   'InnoDB' and
                   table != 'alembic_version']
            self.assertEqual(0, len(res), "%s non InnoDB tables created" % res)


class TestSanityCheck(testlib_api.SqlTestCaseLight):
    BUILD_SCHEMA = False

    def setUp(self):
        super().setUp()
        self.alembic_config = migration.get_neutron_config()
        self.alembic_config.neutron_config = cfg.CONF

    def _drop_table(self, table):
        with self.engine.begin() as conn:
            table.drop(conn)

    def test_check_sanity_1df244e556f5(self):
        ha_router_agent_port_bindings = sqlalchemy.Table(
            'ha_router_agent_port_bindings', sqlalchemy.MetaData(),
            sqlalchemy.Column('port_id', sqlalchemy.String(36)),
            sqlalchemy.Column('router_id', sqlalchemy.String(36)),
            sqlalchemy.Column('l3_agent_id', sqlalchemy.String(36)))

        with self.engine.begin() as conn:
            ha_router_agent_port_bindings.create(conn)
            self.addCleanup(self._drop_table, ha_router_agent_port_bindings)
            # NOTE(haleyb): without this disabled, pylint complains
            # about a missing 'dml' argument.
            # pylint: disable=no-value-for-parameter
            conn.execute(ha_router_agent_port_bindings.insert(), [
                {'port_id': '1234', 'router_id': '12345',
                 'l3_agent_id': '123'},
                {'port_id': '12343', 'router_id': '12345',
                 'l3_agent_id': '123'}
            ])
            script_dir = alembic_script.ScriptDirectory.from_config(
                self.alembic_config)
            script = script_dir.get_revision("1df244e556f5").module
            self.assertRaises(script.DuplicateL3HARouterAgentPortBinding,
                              script.check_sanity, conn)

    def test_check_sanity_030a959ceafa(self):
        routerports = sqlalchemy.Table(
            'routerports', sqlalchemy.MetaData(),
            sqlalchemy.Column('router_id', sqlalchemy.String(36)),
            sqlalchemy.Column('port_id', sqlalchemy.String(36)),
            sqlalchemy.Column('port_type', sqlalchemy.String(255)))

        with self.engine.begin() as conn:
            routerports.create(conn)
            self.addCleanup(self._drop_table, routerports)
            # NOTE(haleyb): without this disabled, pylint complains
            # about a missing 'dml' argument.
            # pylint: disable=no-value-for-parameter
            conn.execute(routerports.insert(), [
                {'router_id': '1234', 'port_id': '12345',
                 'port_type': '123'},
                {'router_id': '12343', 'port_id': '12345',
                 'port_type': '1232'}
            ])
            script_dir = alembic_script.ScriptDirectory.from_config(
                self.alembic_config)
            script = script_dir.get_revision("030a959ceafa").module
            self.assertRaises(script.DuplicatePortRecordinRouterPortdatabase,
                              script.check_sanity, conn)

    def test_check_sanity_6b461a21bcfc_dup_on_fixed_ip(self):
        floatingips = sqlalchemy.Table(
            'floatingips', sqlalchemy.MetaData(),
            sqlalchemy.Column('floating_network_id', sqlalchemy.String(36)),
            sqlalchemy.Column('fixed_port_id', sqlalchemy.String(36)),
            sqlalchemy.Column('fixed_ip_address', sqlalchemy.String(64)))

        with self.engine.begin() as conn:
            floatingips.create(conn)
            self.addCleanup(self._drop_table, floatingips)
            # NOTE(haleyb): without this disabled, pylint complains
            # about a missing 'dml' argument.
            # pylint: disable=no-value-for-parameter
            conn.execute(floatingips.insert(), [
                {'floating_network_id': '12345',
                 'fixed_port_id': '1234567',
                 'fixed_ip_address': '12345678'},
                {'floating_network_id': '12345',
                 'fixed_port_id': '1234567',
                 'fixed_ip_address': '12345678'}
            ])
            script_dir = alembic_script.ScriptDirectory.from_config(
                self.alembic_config)
            script = script_dir.get_revision("6b461a21bcfc").module
            self.assertRaises(script.DuplicateFloatingIPforOneFixedIP,
                              script.check_sanity, conn)

    def test_check_sanity_6b461a21bcfc_dup_on_no_fixed_ip(self):
        floatingips = sqlalchemy.Table(
            'floatingips', sqlalchemy.MetaData(),
            sqlalchemy.Column('floating_network_id', sqlalchemy.String(36)),
            sqlalchemy.Column('fixed_port_id', sqlalchemy.String(36)),
            sqlalchemy.Column('fixed_ip_address', sqlalchemy.String(64)))

        with self.engine.begin() as conn:
            floatingips.create(conn)
            self.addCleanup(self._drop_table, floatingips)
            # NOTE(haleyb): without this disabled, pylint complains
            # about a missing 'dml' argument.
            # pylint: disable=no-value-for-parameter
            conn.execute(floatingips.insert(), [
                {'floating_network_id': '12345',
                 'fixed_port_id': '1234567',
                 'fixed_ip_address': None},
                {'floating_network_id': '12345',
                 'fixed_port_id': '1234567',
                 'fixed_ip_address': None}
            ])
            script_dir = alembic_script.ScriptDirectory.from_config(
                self.alembic_config)
            script = script_dir.get_revision("6b461a21bcfc").module
            self.assertIsNone(script.check_sanity(conn))


class TestWalkDowngrade(oslotest_base.BaseTestCase):

    def setUp(self):
        super().setUp()
        self.alembic_config = migration.get_neutron_config()
        self.alembic_config.neutron_config = cfg.CONF

    def test_no_downgrade(self):
        script_dir = alembic_script.ScriptDirectory.from_config(
            self.alembic_config)
        versions = list(script_dir.walk_revisions(base='base', head='heads'))
        failed_revisions = []
        for version in versions:
            if hasattr(version.module, 'downgrade'):
                failed_revisions.append(version.revision)

        if failed_revisions:
            self.fail('Migrations %s have downgrade' % failed_revisions)
            return True


class _BaseTestWalkMigrations:

    BUILD_SCHEMA = False

    def _get_alembic_config(self, uri):
        db_config = migration.get_neutron_config()
        self.script_dir = alembic_script.ScriptDirectory.from_config(db_config)
        db_config.neutron_config = cfg.CONF
        db_config.neutron_config.set_override('connection',
                                              str(uri),
                                              group='database')
        return db_config


class TestWalkMigrations(_BaseTestWalkMigrations,
                         testlib_api.MySQLTestCaseMixin,
                         testlib_api.SqlTestCaseLight):
    '''This will add framework for testing schema migration
       for different backends.

    '''

    def execute_cmd(self, cmd=None):
        with subprocess.Popen(cmd, stdout=subprocess.PIPE,
                              stderr=subprocess.STDOUT, shell=True) as proc:
            output = proc.communicate()[0]
            self.assertEqual(0, proc.returncode, 'Command failed with '
                             'output:\n%s' % output)

    def _revisions(self):
        """Provides revisions and its parent revisions.

        :return: List of tuples. Every tuple contains revision and its parent
        revision.
        """
        revisions = list(self.script_dir.walk_revisions("base", "heads"))
        revisions = list(reversed(revisions))

        for rev in revisions:
            # Destination, current
            yield rev.revision, rev.down_revision

    def _migrate_up(self, config, engine, dest, curr):
        data = None
        check = getattr(self, "_check_%s" % dest, None)
        pre_upgrade = getattr(self, "_pre_upgrade_%s" % dest, None)
        if pre_upgrade:
            if curr:
                migration.do_alembic_command(config, 'upgrade', curr)
            data = pre_upgrade(engine)

        if check and data:
            migration.do_alembic_command(config, 'upgrade', dest)
            check(engine, data)

    # NOTE(slaweq): this workaround is taken from Manila patch:
    # https://review.opendev.org/#/c/291397/
    # Set 5 minutes timeout for case of running it on very slow nodes/VMs.
    # Note, that this test becomes slower with each addition of new DB
    # migration. On fast nodes it can take about 5-10 secs having Mitaka set of
    # migrations.
    @test_base.set_timeout(600)
    def test_walk_versions(self):
        """Test migrations ability to upgrade and downgrade.

        """
        engine = self.engine
        url_str = render_url_str(engine.url)
        config = self._get_alembic_config(url_str)
        revisions = self._revisions()
        upgrade_dest = None
        for dest, curr in revisions:
            self._migrate_up(config, engine, dest, curr)
            upgrade_dest = dest

        if upgrade_dest:
            migration.do_alembic_command(config, 'upgrade', upgrade_dest)


class TestMigrationsIdempotency(_BaseTestWalkMigrations,
                                testlib_api.MySQLTestCaseMixin,
                                testlib_api.SqlTestCaseLight):
    '''This class tests if the migration scripts are idempotent
    '''

    def _revert_alembic_version(self, target_versions=None):
        alembic_version = sqlalchemy.Table(
            'alembic_version', sqlalchemy.MetaData(),
            sqlalchemy.Column('version_num', sqlalchemy.String(32)))

        with self.engine.begin() as conn:
            # Revision "5c85685d616d" is the head of the CONTRACT branch,
            # it is from Newton release and we don't allow any new CONTRACT
            # DB upgrades, so let's don't bother with that branch
            conn.execute(
                alembic_version.delete().where(
                    alembic_version.c.version_num != '5c85685d616d'
                )
            )
            if target_versions:
                conn.execute(
                    alembic_version.insert(),
                    [{'version_num': tv} for tv in target_versions]
                )

    # NOTE(slaweq): this workaround is taken from Manila patch:
    # https://review.opendev.org/#/c/291397/
    # Set 5 minutes timeout for case of running it on very slow nodes/VMs.
    # Note, that this test becomes slower with each addition of new DB
    # migration. On fast nodes it can take about 5-10 secs having Mitaka set of
    # migrations.
    @test_base.set_timeout(600)
    def test_db_upgrade_is_idempotent(self):
        """Tests if Alembic upgrade scripts are idempotent.

        This function tests if running Alembic upgrade scripts multiple times
        results in the same database state. It does this by first upgrading the
        database to the latest revision, then reverting it to a previous state,
        and finally upgrading it again to the latest revision.
        """
        url_str = render_url_str(self.engine.url)
        config = self._get_alembic_config(url_str)
        migration.do_alembic_command(config, 'upgrade', 'heads')

        # Now lets get back with revision to the 2023.2 HEAD ('89c58a70ceba')
        # and then test again upgrade from from that point through all next
        # releases, starting from 2024.1 if db migration scripts are idempotent
        self._revert_alembic_version(["89c58a70ceba"])
        migration.do_alembic_command(config, 'upgrade', 'heads')



# Copyright (c) 2023 China Unicom Cloud Data Co.,Ltd.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from neutron_lib import constants
from neutron_lib.plugins.ml2 import ovs_constants as p_const
from oslo_log import log as logging

from os_ken.lib.packet import arp
from os_ken.lib.packet import ether_types
from os_ken.lib.packet import in_proto

LOG = logging.getLogger(__name__)
METADATA_V4_IP = constants.METADATA_V4_IP
METADATA_V4_PORT = constants.METADATA_PORT


class MetadataDataPathFlows:

    def set_path_br(self, path_br):
        self.path_br = path_br

    def metadata_path_defaults(self, pvid, to_int_ofport, metadata_ofport,
                               metadata_host_info):
        for table in [p_const.METADATA_EGRESS_NAT,
                      p_const.METADATA_IP_ARP_RESPONDER,
                      p_const.METADATA_INGRESS_DIRECT]:
            self.path_br.install_drop(table_id=table)
        self.metadata_egress_direct(to_int_ofport)
        self.metadata_ingress_direct(pvid, metadata_ofport)
        self.metadata_path_classify(metadata_ofport,
                                    metadata_host_info)
        self.metadata_arp_direct(metadata_ofport)

    def metadata_egress_direct(self, in_port):
        self.path_br.install_goto(table_id=p_const.LOCAL_SWITCHING,
                                  priority=201,
                                  in_port=in_port,
                                  ipv4_dst=METADATA_V4_IP,
                                  eth_type=ether_types.ETH_TYPE_IP,
                                  dest_table_id=p_const.METADATA_EGRESS_NAT)

    def metadata_ingress_direct(self, pvid, in_port):
        (_dp, ofp, ofpp) = self.path_br._get_dp()
        match = ofpp.OFPMatch(in_port=in_port,
                              eth_type=ether_types.ETH_TYPE_IP)

        actions = [
            ofpp.OFPActionPushVlan(),
            ofpp.OFPActionSetField(vlan_vid=pvid | ofp.OFPVID_PRESENT)
        ]
        instructions = [
            ofpp.OFPInstructionActions(ofp.OFPIT_APPLY_ACTIONS, actions),
            ofpp.OFPInstructionGotoTable(
                table_id=p_const.METADATA_INGRESS_DIRECT)
        ]
        self.path_br.install_instructions(
            table_id=p_const.LOCAL_SWITCHING,
            priority=201,
            instructions=instructions,
            match=match)

    def metadata_arp_direct(self, in_port):
        (_dp, ofp, ofpp) = self.path_br._get_dp()
        match = ofpp.OFPMatch(in_port=in_port,
                              eth_type=ether_types.ETH_TYPE_ARP)

        instructions = [
            ofpp.OFPInstructionGotoTable(
                table_id=p_const.METADATA_IP_ARP_RESPONDER)
        ]

        self.path_br.install_instructions(
            table_id=p_const.LOCAL_SWITCHING,
            priority=201,
            instructions=instructions,
            match=match)

    def metadata_path_classify(self, metadata_ofport,
                               metadata_host_info):
        agent_metadata_ip = metadata_host_info.get("provider_ip")
        agent_metadata_mac = metadata_host_info.get("mac_address")
        listen_port = metadata_host_info.get("service_protocol_port")

        (_dp, ofp, ofpp) = self.path_br._get_dp()
        match = ofpp.OFPMatch(
            eth_type=ether_types.ETH_TYPE_IP,
            ipv4_dst=METADATA_V4_IP,
            tcp_dst=METADATA_V4_PORT,
            ip_proto=in_proto.IPPROTO_TCP)
        actions = [
            ofpp.OFPActionSetField(eth_dst=agent_metadata_mac),
            ofpp.OFPActionSetField(ipv4_dst=agent_metadata_ip),
            ofpp.OFPActionSetField(tcp_dst=listen_port),
            ofpp.OFPActionOutput(metadata_ofport, 0),
        ]
        instructions = [
            ofpp.OFPInstructionActions(ofp.OFPIT_APPLY_ACTIONS, actions),
        ]
        self.path_br.install_instructions(
            table_id=p_const.METADATA_EGRESS_OUTPUT,
            priority=202,
            match=match,
            instructions=instructions)

    @staticmethod
    def get_arp_responder_match(ofp, ofpp, ip, in_port=None):
        if in_port:
            match = ofpp.OFPMatch(in_port=in_port,
                                  eth_type=ether_types.ETH_TYPE_ARP,
                                  arp_op=arp.ARP_REQUEST,
                                  arp_tpa=ip)
        else:
            match = ofpp.OFPMatch(eth_type=ether_types.ETH_TYPE_ARP,
                                  arp_op=arp.ARP_REQUEST,
                                  arp_tpa=ip)
        return match

    def install_arp_responder(self, ip, mac, bridge=None, table=None,
                              in_port=None):
        br = bridge if bridge else self.path_br
        table_id = (table if table is not None else
                    p_const.METADATA_IP_ARP_RESPONDER)
        (_dp, ofp, ofpp) = br._get_dp()
        match = self.get_arp_responder_match(ofp, ofpp, ip, in_port)
        actions = [ofpp.OFPActionSetField(arp_op=arp.ARP_REPLY),
                   ofpp.NXActionRegMove(src_field='arp_sha',
                                        dst_field='arp_tha',
                                        n_bits=48),
                   ofpp.NXActionRegMove(src_field='arp_spa',
                                        dst_field='arp_tpa',
                                        n_bits=32),
                   ofpp.OFPActionSetField(arp_sha=mac),
                   ofpp.OFPActionSetField(arp_spa=ip),
                   ofpp.NXActionRegMove(src_field='eth_src',
                                        dst_field='eth_dst',
                                        n_bits=48),
                   ofpp.OFPActionSetField(eth_src=mac),
                   ofpp.OFPActionOutput(ofp.OFPP_IN_PORT, 0)]
        br.install_apply_actions(
            table_id=table_id,
            priority=200,
            match=match,
            actions=actions)

    def delete_arp_responder(self, ip, bridge=None, table=None,
                             in_port=None):
        br = bridge if bridge else self.path_br
        table_id = (table if table is not None else
                    p_const.METADATA_IP_ARP_RESPONDER)
        (_dp, ofp, ofpp) = br._get_dp()
        match = self.get_arp_responder_match(ofp, ofpp, ip, in_port)
        br.uninstall_flows(table_id=table_id, match=match)

    def add_flow_snat_br_meta(self, vlan, src_mac, src_ip,
                              provider_mac, provider_ip):
        (_dp, ofp, ofpp) = self.path_br._get_dp()

        match = ofpp.OFPMatch(vlan_vid=vlan | ofp.OFPVID_PRESENT,
                              eth_type=ether_types.ETH_TYPE_IP,
                              eth_src=src_mac,
                              ipv4_src=src_ip)
        actions = [
            ofpp.OFPActionPopVlan(),
            ofpp.OFPActionSetField(eth_src=provider_mac),
            ofpp.OFPActionSetField(ipv4_src=provider_ip),
        ]
        instructions = [
            ofpp.OFPInstructionActions(ofp.OFPIT_APPLY_ACTIONS, actions),
            ofpp.OFPInstructionGotoTable(
                table_id=p_const.METADATA_EGRESS_OUTPUT)
        ]
        self.path_br.install_instructions(table_id=p_const.METADATA_EGRESS_NAT,
                                          priority=201,
                                          instructions=instructions,
                                          match=match)

    def add_flow_ingress_dnat_direct_to_int_br(
            self, vlan, provider_vlan, provider_ip,
            dst_mac, dst_ip, patch_ofport,
            tap_meta_ofport):
        (_dp, ofp, ofpp) = self.path_br._get_dp()

        match = ofpp.OFPMatch(
            vlan_vid=provider_vlan | ofp.OFPVID_PRESENT,
            in_port=tap_meta_ofport,
            eth_type=ether_types.ETH_TYPE_IP,
            ipv4_dst=provider_ip,
            ip_proto=in_proto.IPPROTO_TCP)
        actions = [
            ofpp.OFPActionSetField(vlan_vid=vlan | ofp.OFPVID_PRESENT),
            ofpp.OFPActionSetField(eth_dst=dst_mac),
            ofpp.OFPActionSetField(
                ipv4_src=METADATA_V4_IP),
            ofpp.OFPActionSetField(ipv4_dst=dst_ip),
            ofpp.OFPActionSetField(
                tcp_src=METADATA_V4_PORT),
            ofpp.OFPActionOutput(patch_ofport, 0),
        ]
        instructions = [
            ofpp.OFPInstructionActions(ofp.OFPIT_APPLY_ACTIONS, actions),
        ]
        self.path_br.install_instructions(
            table_id=p_const.METADATA_INGRESS_DIRECT,
            priority=202,
            instructions=instructions,
            match=match)

    def add_flow_int_br_egress_direct(self, in_port, vlan, patch_ofport):
        (_dp, ofp, ofpp) = self.int_br._get_dp()

        match = ofpp.OFPMatch(
            in_port=in_port,
            eth_type=ether_types.ETH_TYPE_IP,
            ipv4_dst=METADATA_V4_IP)

        actions = [
            ofpp.OFPActionPushVlan(),
            ofpp.OFPActionSetField(vlan_vid=vlan | ofp.OFPVID_PRESENT),
            ofpp.OFPActionOutput(patch_ofport, 0),
        ]
        instructions = [
            ofpp.OFPInstructionActions(ofp.OFPIT_APPLY_ACTIONS, actions),
        ]
        self.int_br.install_instructions(table_id=p_const.LOCAL_SWITCHING,
                                         priority=200,
                                         instructions=instructions,
                                         match=match)

    def add_flow_int_br_ingress_output(self, in_port, vlan, mac, vm_ofport):
        (_dp, ofp, ofpp) = self.int_br._get_dp()

        match = ofpp.OFPMatch(in_port=in_port,
                              eth_type=ether_types.ETH_TYPE_IP,
                              vlan_vid=vlan | ofp.OFPVID_PRESENT,
                              eth_dst=mac,
                              ipv4_src=METADATA_V4_IP)
        actions = [
            ofpp.OFPActionPopVlan(),
            ofpp.OFPActionOutput(vm_ofport, 0),
        ]
        instructions = [
            ofpp.OFPInstructionActions(ofp.OFPIT_APPLY_ACTIONS, actions)
        ]
        self.int_br.install_instructions(table_id=p_const.LOCAL_SWITCHING,
                                         priority=200,
                                         instructions=instructions,
                                         match=match)

    def remove_port_metadata_path_nat_and_arp_flow(
            self, port_vlan, port_mac, port_fixed_ip,
            provider_vlan, provider_ip_addr):
        (_dp, ofp, _ofpp) = self.path_br._get_dp()

        self.path_br.uninstall_flows(
            table_id=p_const.METADATA_EGRESS_NAT,
            eth_type=ether_types.ETH_TYPE_IP,
            eth_src=port_mac,
            ipv4_src=port_fixed_ip,
            vlan_vid=port_vlan | ofp.OFPVID_PRESENT)
        self.path_br.uninstall_flows(
            table_id=p_const.METADATA_IP_ARP_RESPONDER,
            eth_type=ether_types.ETH_TYPE_ARP,
            arp_tpa=provider_ip_addr)
        self.path_br.uninstall_flows(
            table_id=p_const.METADATA_INGRESS_DIRECT,
            eth_type=ether_types.ETH_TYPE_IP,
            ipv4_dst=provider_ip_addr,
            vlan_vid=provider_vlan | ofp.OFPVID_PRESENT)

    def remove_port_metadata_direct_flow(
            self, ofport, port_vlan, port_mac,
            ofport_int_to_snat):
        (_dp, ofp, _ofpp) = self.int_br._get_dp()

        self.int_br.uninstall_flows(
            table_id=p_const.LOCAL_SWITCHING,
            eth_type=ether_types.ETH_TYPE_IP,
            ipv4_dst=METADATA_V4_IP,
            in_port=ofport)
        self.int_br.uninstall_flows(
            table_id=p_const.LOCAL_SWITCHING,
            eth_type=ether_types.ETH_TYPE_IP,
            ipv4_src=METADATA_V4_IP,
            in_port=ofport_int_to_snat,
            vlan_vid=port_vlan | ofp.OFPVID_PRESENT,
            eth_dst=port_mac)
